/**
 *   Copyright (c) Rich Hickey. All rights reserved.
 *   The use and distribution terms for this software are covered by the
 *   Eclipse Public License 1.0 (http://opensource.org/licenses/eclipse-1.0.php)
 *   which can be found in the file epl-v10.html at the root of this distribution.
 *   By using this software in any fashion, you are agreeing to be bound by
 * 	 the terms of this license.
 *   You must not remove this notice, or any other, from this software.
 **/

package com.urbancode.commons.util.immutable;

import java.util.AbstractCollection;
import java.util.AbstractSet;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

abstract class AImmutableMap<K, V> implements IImmutableMap<K, V> {
    int _hash = -1;

    protected AImmutableMap() {
    }

    @Override
    public String toString() {
        Iterator<Entry<K, V>> i = entrySet().iterator();
        if (!i.hasNext()) {
            return "{}";
        }

        StringBuilder builder = new StringBuilder();
        builder.append('{');
        for (;;) {
            Entry<K, V> e = i.next();
            K key = e.getKey();
            V value = e.getValue();
            builder.append(key == this ? "(this Map)" : key);
            builder.append('=');
            builder.append(value == this ? "(this Map)" : value);
            if (!i.hasNext()) {
                return builder.append('}').toString();
            }
            builder.append(", ");
        }
    }

    @Override
    public boolean equals(Object obj) {
        if (!(obj instanceof Map<?, ?>)) {
            return false;
        }
        Map<?, ?> m = (Map<?, ?>) obj;

        if (m.size() != size() || m.hashCode() != hashCode()) {
            return false;
        }

        for (Map.Entry<K, V> e : entrySet()) {
            boolean found = m.containsKey(e.getKey());

            if (!found || !Util.equals(e.getValue(), m.get(e.getKey()))) {
                return false;
            }
        }

        return true;
    }

    @Override
    public int hashCode() {
        if (_hash == -1) {
            int hash = 0;
            for (Map.Entry<K, V> e : entrySet()) {
                hash += (e.getKey() == null ? 0 : e.getKey().hashCode())
                        ^ (e.getValue() == null ? 0 : e.getValue().hashCode());
            }
            this._hash = hash;
        }
        return _hash;
    }

    // java.util.Map implementation

    public void clear() {
        throw new UnsupportedOperationException();
    }

    public boolean containsValue(Object value) {
        return values().contains(value);
    }

    public Set<Entry<K, V>> entrySet() {
        return new AbstractSet<Entry<K, V>>() {

            @Override
            public Iterator<Entry<K, V>> iterator() {
                return AImmutableMap.this.iterator();
            }

            @Override
            public int size() {
                return AImmutableMap.this.size();
            }

            @Override
            public int hashCode() {
                return AImmutableMap.this.hashCode();
            }

            @Override
            public boolean contains(Object o) {
                if (o instanceof Entry<?, ?>) {
                    Entry<?, ?> e = (Entry<?, ?>) o;
                    Entry<?, ?> found = entryAt(e.getKey());
                    if (found != null && Util.equals(found.getValue(), e.getValue())) {
                        return true;
                    }
                }
                return false;
            }
        };
    }

    abstract public V get(Object key);

    public boolean isEmpty() {
        return size() == 0;
    }

    public Set<K> keySet() {
        return new AbstractSet<K>() {

            @Override
            public Iterator<K> iterator() {
                final Iterator<Entry<K, V>> mi = AImmutableMap.this.iterator();

                return new Iterator<K>() {

                    public boolean hasNext() {
                        return mi.hasNext();
                    }

                    public K next() {
                        return mi.next().getKey();
                    }

                    public void remove() {
                        throw new UnsupportedOperationException();
                    }
                };
            }

            @Override
            public int size() {
                return AImmutableMap.this.size();
            }

            @Override
            public boolean contains(Object o) {
                return AImmutableMap.this.containsKey(o);
            }
        };
    }

    public V put(K key, V value) {
        throw new UnsupportedOperationException();
    }

    public void putAll(Map<? extends K, ? extends V> t) {
        throw new UnsupportedOperationException();
    }

    public V remove(Object key) {
        throw new UnsupportedOperationException();
    }

    public Collection<V> values() {
        return new AbstractCollection<V>() {

            @Override
            public Iterator<V> iterator() {
                final Iterator<Entry<K, V>> mi = AImmutableMap.this.iterator();

                return new Iterator<V>() {

                    public boolean hasNext() {
                        return mi.hasNext();
                    }

                    public V next() {
                        return mi.next().getValue();
                    }

                    public void remove() {
                        throw new UnsupportedOperationException();
                    }
                };
            }

            @Override
            public int size() {
                return AImmutableMap.this.size();
            }
        };
    }

    Iterator<Map.Entry<K, V>> iterator() {
        return new SeqIterator<Map.Entry<K, V>>(seq());
    }

    abstract protected ISeq<Map.Entry<K, V>> seq();

    abstract protected Map.Entry<K, V> entryAt(Object key);
}
