/**
 *   Copyright (c) Rich Hickey. All rights reserved.
 *   The use and distribution terms for this software are covered by the
 *   Eclipse Public License 1.0 (http://opensource.org/licenses/eclipse-1.0.php)
 *   which can be found in the file epl-v10.html at the root of this distribution.
 *   By using this software in any fashion, you are agreeing to be bound by
 * 	 the terms of this license.
 *   You must not remove this notice, or any other, from this software.
 **/

/* rich Dec 18, 2007 */

package com.urbancode.commons.util.immutable;

import java.util.Collection;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.RandomAccess;

abstract class AImmutableVector<T> implements IImmutableVector<T>, RandomAccess {
    int _hash = -1;

    public AImmutableVector() {
    }

    @Override
    public String toString() {
        return Util.toString(this, iterator());
        }

    @Override
    public boolean equals(Object object) {
        return Util.equals(this, object);
    }

    @Override
    public int hashCode() {
        if (_hash == -1) {
            this._hash = Util.hashCode(this);
            }
        return _hash;
    }

    public int indexOf(Object o) {
        for (int i = 0; i < size(); i++) {
            if (Util.equals(get(i), o)) {
                return i;
            }
        }
        return -1;
    }

    public int lastIndexOf(Object o) {
        for (int i = size() - 1; i >= 0; i--) {
            if (Util.equals(get(i), o)) {
                return i;
            }
        }
        return -1;
    }

    public ListIterator<T> listIterator() {
        return listIterator(0);
    }

    public ListIterator<T> listIterator(final int index) {
        return new ListIterator<T>() {
            int nexti = index;

            public boolean hasNext() {
                return nexti < size();
            }

            public T next() {
                return get(nexti++);
            }

            public boolean hasPrevious() {
                return nexti > 0;
            }

            public T previous() {
                return get(--nexti);
            }

            public int nextIndex() {
                return nexti;
            }

            public int previousIndex() {
                return nexti - 1;
            }

            public void remove() {
                throw new UnsupportedOperationException();
            }

            public void set(Object o) {
                throw new UnsupportedOperationException();
            }

            public void add(Object o) {
                throw new UnsupportedOperationException();
            }
        };
    }

    public IImmutableVector<T> subList(int fromIndex, int toIndex) {
        if (toIndex < fromIndex || fromIndex < 0 || toIndex > size()) {
            throw new IndexOutOfBoundsException();
        }
        if (fromIndex == toIndex) {
            return ImmutableVector.empty();
        }
        return new AImmutableVector.SubVector<T>(this, fromIndex, toIndex);
    }

    public Iterator<T> iterator() {
        // todo - something more efficient
        return new Iterator<T>() {
            int i = 0;

            public boolean hasNext() {
                return i < size();
            }

            public T next() {
                return get(i++);
            }

            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }

    public T peek() {
        if (size() > 0) {
            return get(size() - 1);
        }
        return null;
    }

    // java.util.Collection implementation

    public boolean containsAll(Collection<?> c) {
        for (Object o : c) {
            if (!contains(o)) {
                return false;
            }
        }
        return true;
    }

    public <E> E[] toArray(E[] array) {
        return Util.toArray(size(), iterator(), array);
        }

    public Object[] toArray() {
        return Util.toArray(size(), iterator());
        }

    public boolean isEmpty() {
        return size() == 0;
    }

    public boolean contains(Object o) {
        for (T t : this) {
            if (Util.equals(t, o)) {
                return true;
            }
        }
        return false;
    }

    public T set(int i, T o) {
        throw new UnsupportedOperationException();
    }

    public void add(int i, Object o) {
        throw new UnsupportedOperationException();
    }

    public boolean addAll(int i, Collection<? extends T> c) {
        throw new UnsupportedOperationException();
    }

    public boolean add(T o) {
        throw new UnsupportedOperationException();
    }

    public boolean remove(Object o) {
        throw new UnsupportedOperationException();
    }

    public boolean addAll(Collection<? extends T> c) {
        throw new UnsupportedOperationException();
    }

    public void clear() {
        throw new UnsupportedOperationException();
    }

    public boolean retainAll(Collection<?> c) {
        throw new UnsupportedOperationException();
    }

    public T remove(int i) {
        throw new UnsupportedOperationException();
    }

    public boolean removeAll(Collection<?> c) {
        throw new UnsupportedOperationException();
    }

    static class SubVector<T> extends AImmutableVector<T> {
        final IImmutableVector<T> v;
        final int start;
        final int end;

        public SubVector(IImmutableVector<T> v, int start, int end) {
            if (v instanceof AImmutableVector.SubVector<?>) {
                AImmutableVector.SubVector<T> sv = (AImmutableVector.SubVector<T>) v;
                start += sv.start;
                end += sv.start;
                v = sv.v;
            }
            this.v = v;
            this.start = start;
            this.end = end;
        }

        public T get(int i) {
            if (start + i >= end) {
                throw new IndexOutOfBoundsException();
            }
            return v.get(start + i);
        }

        public IImmutableVector<T> with(int i, T val) {
            if (start + i > end) {
                throw new IndexOutOfBoundsException();
            }
            else if (start + i == end) {
                return with(val);
            }
            return new SubVector<T>(v.with(start + i, val), start, end);
        }

        public int size() {
            return end - start;
        }

        public IImmutableVector<T> pop() {
            if (end - 1 == start) {
                return ImmutableVector.empty();
            }
            return new SubVector<T>(v, start, end - 1);
        }

        public IImmutableVector<T> with(T o) {
            return new SubVector<T>(v.with(end, o), start, end + 1);
        }
    }
}
