/**
 Copyright (c) 2007-2008, Rich Hickey
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 * Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.

 * Neither the name of Clojure nor the names of its contributors
   may be used to endorse or promote products derived from this
   software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.
 **/

/* rich Jun 18, 2007 */

package com.urbancode.commons.util.immutable;

import java.util.Map;

//this stuff is just for the test main()

/*
 A persistent rendition of Phil Bagwell's Hash Array Mapped Trie

 Uses path copying for persistence
 HashCollision leaves vs. extended hashing
 Node polymorphism vs. conditionals
 No sub-tree pools or root-resizing
 Any errors are my own
 */

@SuppressWarnings("unchecked")
public class ImmutableHashMap<K, V> extends AImmutableMap<K, V> {

    final int count;
    final INode<K, V> root;

    final static private EmptyNode<Object, Object> EMPTY_NODE = new EmptyNode<Object, Object>();
    final static private ImmutableHashMap<?, ?> EMPTY = new ImmutableHashMap<Object, Object>(0, EMPTY_NODE);

    static public <K, V> ImmutableHashMap<K, V> empty() {
        return (ImmutableHashMap<K, V>) EMPTY;
    }

    static public <K, V> ImmutableHashMap<K, V> create(Map<? extends K, ? extends V> other) {
        ImmutableHashMap<K, V> ret = empty();
        for (Entry<? extends K, ? extends V> o : other.entrySet()) {
            ret = ret.with(o.getKey(), o.getValue());
        }
        return ret;
    }

    private ImmutableHashMap(int count, INode<K, V> root) {
        this.count = count;
        this.root = root;
    }

    public boolean containsKey(Object key) {
        return entryAt(key) != null;
    }

    @Override
    protected Entry<K, V> entryAt(Object key) {
        return root.find(Util.hash(key), key);
    }

    public ImmutableHashMap<K, V> with(K key, V val) {
        Box addedLeaf = new Box(null);
        INode<K, V> newroot = root.with(0, Util.hash(key), key, val, addedLeaf);
        if (newroot == root) {
            return this;
        }
        return new ImmutableHashMap<K, V>(addedLeaf.val == null ? count : count + 1, newroot);
    }

    public V get(Object key, V notFound) {
        Entry<K, V> e = entryAt(key);
        if (e != null) {
            return e.getValue();
        }
        return notFound;
    }

    @Override
    public V get(Object key) {
        return get(key, null);
    }

    public ImmutableHashMap<K, V> without(Object key) {
        INode<K, V> newroot = root.without(Util.hash(key), key);
        if (newroot == root) {
            return this;
        }
        if (newroot == null) {
            return empty();
        }
        return new ImmutableHashMap<K, V>(count - 1, newroot);
    }

    public int size() {
        return count;
    }

    @Override
    protected ISeq<Map.Entry<K, V>> seq() {
        return root.nodeSeq();
    }

    static int mask(int hash, int shift) {
        return (hash >>> shift) & 0x01f;
    }

    static interface INode<K, V> {
        INode<K, V> with(int shift, int hash, K key, V val, Box addedLeaf);

        INode<K, V> without(int hash, Object key);

        LeafNode<K, V> find(int hash, Object key);

        ISeq<Map.Entry<K, V>> nodeSeq();

        int getHash();
    }

    /*
     * static interface ILeaf extends INode{ int getHash(); }
     */

    final static class EmptyNode<K, V> implements INode<K, V> {

        public INode<K, V> with(int shift, int hash, K key, V val, Box addedLeaf) {
            INode<K, V> ret = new LeafNode<K, V>(hash, key, val);
            addedLeaf.val = ret;
            return ret;
        }

        public INode<K, V> without(int hash, Object key) {
            return this;
        }

        public LeafNode<K, V> find(int hash, Object key) {
            return null;
        }

        public ISeq<Map.Entry<K, V>> nodeSeq() {
            return null;
        }

        public int getHash() {
            return 0;
        }
    }

    final static class FullNode<K, V> implements INode<K, V> {
        final INode<K, V>[] nodes;
        final int shift;
        final int _hash;

        static int bitpos(int hash, int shift) {
            return 1 << mask(hash, shift);
        }

        FullNode(INode<K, V>[] nodes, int shift) {
            this.nodes = nodes;
            this.shift = shift;
            this._hash = nodes[0].getHash();
        }

        public INode<K, V> with(int levelShift, int hash, K key, V val, Box addedLeaf) {
            int idx = mask(hash, shift);

            INode<K, V> n = nodes[idx].with(shift + 5, hash, key, val, addedLeaf);
            if (n == nodes[idx]) {
                return this;
            }
            else {
                INode<K, V>[] newnodes = nodes.clone();
                newnodes[idx] = n;
                return new FullNode<K, V>(newnodes, shift);
            }
        }

        public INode<K, V> without(int hash, Object key) {
            int idx = mask(hash, shift);
            INode<K, V> n = nodes[idx].without(hash, key);
            if (n != nodes[idx]) {
                if (n == null) {
                    INode<K, V>[] newnodes = new INode[nodes.length - 1];
                    System.arraycopy(nodes, 0, newnodes, 0, idx);
                    System.arraycopy(nodes, idx + 1, newnodes, idx, nodes.length - (idx + 1));
                    return new BitmapIndexedNode<K, V>(~bitpos(hash, shift), newnodes, shift);
                }
                INode<K, V>[] newnodes = nodes.clone();
                newnodes[idx] = n;
                return new FullNode<K, V>(newnodes, shift);
            }
            return this;
        }

        public LeafNode<K, V> find(int hash, Object key) {
            return nodes[mask(hash, shift)].find(hash, key);
        }

        public ISeq<Map.Entry<K, V>> nodeSeq() {
            return Seq.create(this, 0);
        }

        public int getHash() {
            return _hash;
        }

        static class Seq<K, V> implements ISeq<Map.Entry<K, V>> {
            final ISeq<? extends Map.Entry<K, V>> s;
            final int i;
            final FullNode<K, V> node;

            Seq(ISeq<? extends Map.Entry<K, V>> s, int i, FullNode<K, V> node) {
                this.s = s;
                this.i = i;
                this.node = node;
            }

            static <K, V> ISeq<Map.Entry<K, V>> create(FullNode<K, V> node, int i) {
                if (i >= node.nodes.length) {
                    return null;
                }
                return new Seq<K, V>(node.nodes[i].nodeSeq(), i, node);
            }

            public Map.Entry<K, V> first() {
                return s.first();
            }

            public ISeq<Map.Entry<K, V>> next() {
                ISeq<? extends Map.Entry<K, V>> nexts = s.next();
                if (nexts != null) {
                    return new Seq<K, V>(nexts, i, node);
                }
                return create(node, i + 1);
            }
        }
    }

    final static class BitmapIndexedNode<K, V> implements INode<K, V> {
        final int bitmap;
        final INode<K, V>[] nodes;
        final int shift;
        final int _hash;

        static int bitpos(int hash, int shift) {
            return 1 << mask(hash, shift);
        }

        final int index(int bit) {
            return Integer.bitCount(bitmap & (bit - 1));
        }

        BitmapIndexedNode(int bitmap, INode<K, V>[] nodes, int shift) {
            this.bitmap = bitmap;
            this.nodes = nodes;
            this.shift = shift;
            this._hash = nodes[0].getHash();
        }

        static <K, V> INode<K, V> create(int bitmap, INode<K, V>[] nodes, int shift) {
            if (bitmap == -1) {
                return new FullNode<K, V>(nodes, shift);
            }
            return new BitmapIndexedNode<K, V>(bitmap, nodes, shift);
        }

        static <K, V> INode<K, V> create(int shift, INode<K, V> branch, int hash, K key, V val, Box addedLeaf) {
            int position = bitpos(branch.getHash(), shift);
            INode<K, V> node = new BitmapIndexedNode<K, V>(position, new INode[] { branch }, shift);
            return node.with(shift, hash, key, val, addedLeaf);
        }

        public INode<K, V> with(int levelShift, int hash, K key, V val, Box addedLeaf) {
            int bit = bitpos(hash, shift);
            int idx = index(bit);
            if ((bitmap & bit) != 0) {
                INode<K, V> n = nodes[idx].with(shift + 5, hash, key, val, addedLeaf);
                if (n == nodes[idx]) {
                    return this;
                }
                else {
                    INode<K, V>[] newnodes = nodes.clone();
                    newnodes[idx] = n;
                    return new BitmapIndexedNode<K, V>(bitmap, newnodes, shift);
                }
            }
            else {
                INode<K, V>[] newnodes = new INode[nodes.length + 1];
                System.arraycopy(nodes, 0, newnodes, 0, idx);
                addedLeaf.val = newnodes[idx] = new LeafNode<K, V>(hash, key, val);
                System.arraycopy(nodes, idx, newnodes, idx + 1, nodes.length - idx);
                return create(bitmap | bit, newnodes, shift);
            }
        }

        public INode<K, V> without(int hash, Object key) {
            int bit = bitpos(hash, shift);
            if ((bitmap & bit) != 0) {
                int idx = index(bit);
                INode<K, V> n = nodes[idx].without(hash, key);
                if (n != nodes[idx]) {
                    if (n == null) {
                        if (bitmap == bit) {
                            return null;
                        }
                        INode<K, V>[] newnodes = new INode[nodes.length - 1];
                        System.arraycopy(nodes, 0, newnodes, 0, idx);
                        System.arraycopy(nodes, idx + 1, newnodes, idx, nodes.length - (idx + 1));
                        return new BitmapIndexedNode<K, V>(bitmap & ~bit, newnodes, shift);
                    }
                    INode<K, V>[] newnodes = nodes.clone();
                    newnodes[idx] = n;
                    return new BitmapIndexedNode<K, V>(bitmap, newnodes, shift);
                }
            }
            return this;
        }

        public LeafNode<K, V> find(int hash, Object key) {
            int bit = bitpos(hash, shift);
            if ((bitmap & bit) != 0) {
                return nodes[index(bit)].find(hash, key);
            }
            else {
                return null;
            }
        }

        public int getHash() {
            return _hash;
        }

        public ISeq<Map.Entry<K, V>> nodeSeq() {
            return Seq.create(this, 0);
        }

        static class Seq<K, V> implements ISeq<Map.Entry<K, V>> {
            final ISeq<? extends Map.Entry<K, V>> s;
            final int i;
            final BitmapIndexedNode<K, V> node;

            Seq(ISeq<? extends Map.Entry<K, V>> s, int i, BitmapIndexedNode<K, V> node) {
                this.s = s;
                this.i = i;
                this.node = node;
            }

            static <K, V> ISeq<Map.Entry<K, V>> create(BitmapIndexedNode<K, V> node, int i) {
                if (i >= node.nodes.length) {
                    return null;
                }
                return new Seq<K, V>(node.nodes[i].nodeSeq(), i, node);
            }

            public Map.Entry<K, V> first() {
                return s.first();
            }

            public ISeq<Map.Entry<K, V>> next() {
                ISeq<? extends Map.Entry<K, V>> nexts = s.next();
                if (nexts != null) {
                    return new Seq<K, V>(nexts, i, node);
                }
                return create(node, i + 1);
            }
        }
    }

    final static class LeafNode<K, V> implements Map.Entry<K, V>, INode<K, V> {
        final int hash;
        final K key;
        final V val;

        public LeafNode(int hash, K key, V val) {
            this.hash = hash;
            this.key = key;
            this.val = val;
        }

        public INode<K, V> with(int shift, int hash, K key, V val, Box addedLeaf) {
            if (hash == this.hash) {
                if (Util.equals(key, this.key)) {
                    if (val == this.val) {
                        return this;
                    }
                    // note - do not set addedLeaf, since we are replacing
                    return new LeafNode<K, V>(hash, key, val);
                }
                // hash collision - same hash, different keys
                LeafNode<K, V> newLeaf = new LeafNode<K, V>(hash, key, val);
                addedLeaf.val = newLeaf;
                return new HashCollisionNode<K, V>(hash, this, newLeaf);
            }
            return BitmapIndexedNode.create(shift, this, hash, key, val, addedLeaf);
        }

        public INode<K, V> without(int hash, Object key) {
            if (hash == this.hash && Util.equals(key, this.key)) {
                return null;
            }
            return this;
        }

        public LeafNode<K, V> find(int hash, Object key) {
            if (hash == this.hash && Util.equals(key, this.key)) {
                return this;
            }
            return null;
        }

        public ISeq<Map.Entry<K, V>> nodeSeq() {
            return new ISeq<Entry<K, V>>() {
                public Map.Entry<K, V> first() {
                    return LeafNode.this;
                }

                public ISeq<Map.Entry<K, V>> next() {
                    return null;
                }
            };
        }

        public int getHash() {
            return hash;
        }

        public K key() {
            return this.key;
        }

        public V val() {
            return this.val;
        }

        public K getKey() {
            return this.key;
        }

        public V getValue() {
            return this.val;
        }

        public void setKey(K key) {
            throw new UnsupportedOperationException();
        }

        public V setValue(V value) {
            throw new UnsupportedOperationException();
        }
    }

    final static class HashCollisionNode<K, V> implements INode<K, V> {

        final int hash;
        final LeafNode<K, V>[] leaves;

        public HashCollisionNode(int hash, LeafNode<K, V>... leaves) {
            this.hash = hash;
            this.leaves = leaves;
        }

        public HashCollisionNode(int hash, LeafNode<K, V> leaf) {
            this.hash = hash;
            this.leaves = new LeafNode[] { leaf };
        }

        public INode<K, V> with(int shift, int hash, K key, V val, Box addedLeaf) {
            if (hash == this.hash) {
                int idx = findIndex(hash, key);
                if (idx != -1) {
                    if (leaves[idx].val == val) {
                        return this;
                    }
                    LeafNode<K, V>[] newLeaves = leaves.clone();
                    // note - do not set addedLeaf, since we are replacing
                    newLeaves[idx] = new LeafNode<K, V>(hash, key, val);
                    return new HashCollisionNode<K, V>(hash, newLeaves);
                }
                LeafNode<K, V>[] newLeaves = new LeafNode[leaves.length + 1];
                System.arraycopy(leaves, 0, newLeaves, 0, leaves.length);
                addedLeaf.val = newLeaves[leaves.length] = new LeafNode<K, V>(hash, key, val);
                return new HashCollisionNode<K, V>(hash, newLeaves);
            }
            return BitmapIndexedNode.create(shift, this, hash, key, val, addedLeaf);
        }

        public INode<K, V> without(int hash, Object key) {
            int idx = findIndex(hash, key);
            if (idx == -1) {
                return this;
            }
            if (leaves.length == 2) {
                return idx == 0 ? leaves[1] : leaves[0];
            }
            LeafNode<K, V>[] newLeaves = new LeafNode[leaves.length - 1];
            System.arraycopy(leaves, 0, newLeaves, 0, idx);
            System.arraycopy(leaves, idx + 1, newLeaves, idx, leaves.length - (idx + 1));
            return new HashCollisionNode<K, V>(hash, newLeaves);
        }

        public LeafNode<K, V> find(int hash, Object key) {
            int idx = findIndex(hash, key);
            if (idx != -1) {
                return leaves[idx];
            }
            return null;
        }

        @SuppressWarnings("rawtypes")
        public ISeq<Map.Entry<K, V>> nodeSeq() {
            return (ISeq) ArraySeq.create(leaves);
        }

        int findIndex(int hash, Object key) {
            for (int i = 0; i < leaves.length; i++) {
                if (leaves[i].find(hash, key) != null) {
                    return i;
                }
            }
            return -1;
        }

        public int getHash() {
            return hash;
        }
    }
}
