/**
 *   Copyright (c) Rich Hickey. All rights reserved.
 *   The use and distribution terms for this software are covered by the
 *   Eclipse Public License 1.0 (http://opensource.org/licenses/eclipse-1.0.php)
 *   which can be found in the file epl-v10.html at the root of this distribution.
 *   By using this software in any fashion, you are agreeing to be bound by
 * 	 the terms of this license.
 *   You must not remove this notice, or any other, from this software.
 **/

package com.urbancode.commons.util.immutable;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.NoSuchElementException;

public class ImmutableList<T> implements IImmutableList<T> {
    int hash = -1;
    private final T first;
    private final ImmutableList<T> rest;
    private final int size;

    final static private ImmutableList<?> EMPTY = new ImmutableList<Object>();

    @SuppressWarnings("unchecked")
    static public <T> ImmutableList<T> empty() {
        return (ImmutableList<T>) EMPTY;
    }

    ImmutableList() {
        this.first = null;
        this.rest = null;
        size = 0;
    }

    ImmutableList(T first) {
        this.first = first;
        this.rest = null;
        size = 1;
    }

    ImmutableList(T _first, ImmutableList<T> _rest, int _count) {
        this.first = _first;
        this.rest = _rest;
        this.size = _count;
    }

    public static <T> ImmutableList<T> create(List<? extends T> init) {
        ImmutableList<T> ret = empty();
        for (ListIterator<? extends T> i = init.listIterator(init.size()); i.hasPrevious();) {
            ret = ret.with(i.previous());
        }
        return ret;
    }

    public static <T> ImmutableList<T> create(T... init) {
        ImmutableList<T> ret = empty();
        for (int i = init.length - 1; i >= 0; i--) {
            ret = ret.with(init[i]);
        }
        return ret;
    }

    @Override
    public String toString() {
        return Util.toString(this, iterator());
    }

    @Override
    public int hashCode() {
        if (hash == -1) {
            hash = Util.hashCode(this);
        }
        return hash;
    }

    @Override
    public boolean equals(Object object) {
        return Util.equals(this, object);
    }

    public T first() {
        return first;
    }

    public ImmutableList<T> rest() {
        if (size == 1) {
            return null;
        }
        return rest;
    }

    public T peek() {
        return first();
    }

    public ImmutableList<T> pop() {
        if (rest == null) {
            return empty();
        }
        return rest;
    }

    public int size() {
        return size;
    }

    public ImmutableList<T> with(T o) {
        return new ImmutableList<T>(o, this, size + 1);
    }

    public boolean contains(Object o) {
        for (ImmutableList<T> s = this; s != null; s = s.rest()) {
            if (Util.equals(s.first(), o)) {
                return true;
            }
        }
        return false;
    }


    public boolean isEmpty() {
        return size == 0;
    }

    public Iterator<T> iterator() {
        return new ImmutableListIterator<T>(this);
    }

    public <E> E[] toArray(E[] array) {
        return Util.toArray(size(), iterator(), array);
    }

    public Object[] toArray() {
        return Util.toArray(size(), iterator());
    }

    public int indexOf(Object o) {
        ImmutableList<T> s = this;
        for (int i = 0; s != null; s = s.rest(), i++) {
            if (Util.equals(s.first(), o)) {
                return i;
            }
        }
        return -1;
    }

    public T get(int index) {
        ImmutableList<T> seq = this;
        for (int i = 0; i <= index; ++i, seq = seq.rest()) {
            if (i == index) {
                return first();
            }
        }
        throw new IndexOutOfBoundsException();
    }

    public void add(int index, Object element) {
        throw new UnsupportedOperationException();
    }

    public boolean addAll(int index, Collection<? extends T> c) {
        throw new UnsupportedOperationException();
    }

    public boolean add(Object o) {
        throw new UnsupportedOperationException();
    }

    public boolean remove(Object o) {
        throw new UnsupportedOperationException();
    }

    public boolean addAll(Collection<? extends T> c) {
        throw new UnsupportedOperationException();
    }

    public void clear() {
        throw new UnsupportedOperationException();
    }

    public boolean retainAll(Collection<?> c) {
        throw new UnsupportedOperationException();
    }

    public boolean removeAll(Collection<?> c) {
        throw new UnsupportedOperationException();
    }

    public boolean containsAll(Collection<?> c) {
        for (Object o : c) {
            if (!contains(o)) {
                return false;
            }
        }
        return true;
    }

    public ImmutableList<T> subList(int fromIndex, int toIndex) {
        if (fromIndex < 0 || toIndex > size || fromIndex > toIndex) {
            throw new IndexOutOfBoundsException();
        }
        ArrayList<T> accum = new ArrayList<T>();
        ImmutableList<T> seq = this;
        int i;
        for (i = 0; seq != null && i < fromIndex; i++) {
            seq = seq.rest();
        }
        for (;seq != null && i < toIndex; ++i, seq = seq.rest()) {
            accum.add(seq.first());
        }
        return create(accum);
    }

    public T set(int index, T element) {
        throw new UnsupportedOperationException();
    }

    public T remove(int index) {
        throw new UnsupportedOperationException();
    }

    public int lastIndexOf(Object o) {
        return reify().lastIndexOf(o);
    }

    public ListIterator<T> listIterator() {
        return reify().listIterator();
    }

    public ListIterator<T> listIterator(int index) {
        return reify().listIterator(index);
    }

    private List<T> reify() {
        return Collections.unmodifiableList(new ArrayList<T>(this));
    }

    final static private class ImmutableListIterator<T> implements Iterator<T> {

        IImmutableList<T> list;

        public ImmutableListIterator(IImmutableList<T> list) {
            this.list = list;
        }

        public boolean hasNext() {
            return list != null;
        }

        public T next() {
            if (list == null) {
                throw new NoSuchElementException();
            }
            T ret = list.first();
            list = list.rest();
            return ret;
        }

        public void remove() {
            throw new UnsupportedOperationException();
        }
    }
}
