/**
 *   Copyright (c) Rich Hickey. All rights reserved.
 *   The use and distribution terms for this software are covered by the
 *   Eclipse Public License 1.0 (http://opensource.org/licenses/eclipse-1.0.php)
 *   which can be found in the file epl-v10.html at the root of this distribution.
 *   By using this software in any fashion, you are agreeing to be bound by
 * 	 the terms of this license.
 *   You must not remove this notice, or any other, from this software.
 **/

/* rich May 20, 2006 */

package com.urbancode.commons.util.immutable;

import java.util.Comparator;
import java.util.Map;

/**
 * Persistent Red Black Tree Note that instances of this class are constant
 * values i.e. add/remove etc return new values
 * <p/>
 * See Okasaki, Kahrs, Larsen et al
 */

@SuppressWarnings("unchecked")
public class ImmutableTreeMap<K, V> extends AImmutableMap<K, V> {

    public final Comparator<? super K> comp;
    public final Node<K, V> tree;
    public final int _count;

    final static private ImmutableTreeMap<?, ?> EMPTY = new ImmutableTreeMap<Object, Object>();

    static private final Comparator<?> DEFAULT_COMPARATOR = new Comparator<Object>() {
        public int compare(Object o1, Object o2) {
            return ((Comparable<Object>)o1).compareTo(o2);
        }
    };

    static private <K> Comparator<K> defaultComparator() {
        return (Comparator<K>) DEFAULT_COMPARATOR;
    }

    static public <K, V> ImmutableTreeMap<K, V> empty() {
        return (ImmutableTreeMap<K, V>) EMPTY;
    }

    static public <K, V> ImmutableTreeMap<K, V> empty(Comparator<? super K> comp) {
        if (comp != null) {
            return new ImmutableTreeMap<K, V>(comp);
        }
        else {
            return empty();
        }
    }

    static public <K, V> ImmutableTreeMap<K, V> create(Map<? extends K, ? extends V> other) {
        ImmutableTreeMap<K, V> ret = empty();
        for (Map.Entry<? extends K, ? extends V> o : other.entrySet()) {
            ret = ret.with(o.getKey(), o.getValue());
        }
        return ret;
    }

    ImmutableTreeMap() {
        this(defaultComparator());
    }

    ImmutableTreeMap(Comparator<? super K> comp) {
        super();
        this.comp = comp;
        tree = null;
        _count = 0;
    }

    ImmutableTreeMap(Comparator<? super K> comp, Node<K, V> tree, int _count) {
        this.comp = comp;
        this.tree = tree;
        this._count = _count;
    }




    public V get(Object key, V notFound) {
        Node<K, V> n = entryAt(key);
        return (n != null) ? n.getValue() : notFound;
    }

    @Override
    public V get(Object key) {
        return get(key, null);
    }

    public K firstKey() {
        Node<K, V> t = min();
        return t != null ? t.key : null;
    }
    public K lastKey() {
        Node<K, V> t = max();
        return t != null ? t.key : null;
    }

    public boolean containsKey(Object key) {
        return entryAt(key) != null;
    }

    public ImmutableTreeMap<K, V> with(K key, V val) {
        Box found = new Box(null);
        Node<K, V> t = add(tree, key, val, found);
        if (t == null) // null == already contains key
        {
            Node<K, V> foundNode = (Node<K, V>) found.val;
            if (foundNode.getValue() == val) {
                // identity of val, not equals()
                return this;
            }
            return new ImmutableTreeMap<K, V>(comp, replace(tree, key, val), _count);
        }
        return new ImmutableTreeMap<K, V>(comp, t.blacken(), _count + 1);
    }

    public ImmutableTreeMap<K, V> without(Object key) {
        Box found = new Box(null);
        Node<K, V> t = remove(tree, (K) key, found);
        if (t == null) {
            if (found.val == null) {
                return this;
            }
            // empty
            return new ImmutableTreeMap<K, V>(comp);
        }
        return new ImmutableTreeMap<K, V>(comp, t.blacken(), _count - 1);
    }

    @Override
    protected ISeq<Map.Entry<K, V>> seq() {
        if (_count > 0) {
            return Seq.create(tree, true, _count);
        }
        return null;
    }

    public Comparator<? super K> comparator() {
        return comp;
    }

    Node<K, V> min() {
        Node<K, V> t = tree;
        if (t != null) {
            while (t.left() != null) {
                t = t.left();
            }
        }
        return t;
    }

    Node<K, V> max() {
        Node<K, V> t = tree;
        if (t != null) {
            while (t.right() != null) {
                t = t.right();
            }
        }
        return t;
    }


    public int size() {
        return _count;
    }

    @Override
    public Node<K, V> entryAt(Object key) {
        Node<K, V> t = tree;
        while (t != null) {
            int c = compare((K) key, t.key);
            if (c == 0) {
                return t;
            }
            else if (c < 0) {
                t = t.left();
            }
            else {
                t = t.right();
            }
        }
        return t;
    }

    private int compare(K k1, K k2) {
        return comp.compare(k1, k2);
    }

    Node<K, V> add(Node<K, V> t, K key, V val, Box found) {
        if (t == null) {
            if (val == null) {
                return new Red<K, V>(key);
            }
            return new RedVal<K, V>(key, val);
        }
        int c = compare(key, t.key);
        if (c == 0) {
            found.val = t;
            return null;
        }
        Node<K, V> ins = c < 0 ? add(t.left(), key, val, found) : add(t.right(), key, val, found);
        if (ins == null) {
            return null;
        }
        if (c < 0) {
            return t.addLeft(ins);
        }
        return t.addRight(ins);
    }

    Node<K, V> remove(Node<K, V> t, K key, Box found) {
        if (t == null) {
            return null; // not found indicator
        }
        int c = compare(key, t.key);
        if (c == 0) {
            found.val = t;
            return append(t.left(), t.right());
        }
        Node<K, V> del = c < 0 ? remove(t.left(), key, found) : remove(t.right(), key, found);
        if (del == null && found.val == null) {
            return null;
        }
        if (c < 0) {
            if (t.left() instanceof Black<?, ?>) {
                return balanceLeftDel(t.key, t.getValue(), del, t.right());
            }
            else {
                return red(t.key, t.getValue(), del, t.right());
            }
        }
        if (t.right() instanceof Black<?, ?>) {
            return balanceRightDel(t.key, t.getValue(), t.left(), del);
        }
        return red(t.key, t.getValue(), t.left(), del);
    }

    static <K, V> Node<K, V> append(Node<K, V> left, Node<K, V> right) {
        if (left == null) {
            return right;
        }
        else if (right == null) {
            return left;
        }
        else if (left instanceof Red<?, ?>) {
            if (right instanceof Red<?, ?>) {
                Node<K, V> app = append(left.right(), right.left());
                if (app instanceof Red<?, ?>) {
                    return red(
                            app.key,
                            app.getValue(),
                            red(left.key, left.getValue(), left.left(), app.left()),
                            red(right.key, right.getValue(), app.right(), right.right()));
                }
                else {
                    return red(
                            left.key,
                            left.getValue(),
                            left.left(),
                            red(right.key, right.getValue(), app, right.right()));
                }
            }
            else {
                return red(left.key, left.getValue(), left.left(), append(left.right(), right));
            }
        } else if (right instanceof Red<?, ?>) {
            return red(right.key, right.getValue(), append(left, right.left()), right.right());
        }
        else // black/black
        {
            Node<K, V> app = append(left.right(), right.left());
            if (app instanceof Red<?, ?>) {
                return red(
                        app.key,
                        app.getValue(),
                        black(left.key, left.getValue(), left.left(), app.left()),
                        black(right.key, right.getValue(), app.right(), right.right()));
            }
            else {
                return balanceLeftDel(
                        left.key,
                        left.getValue(),
                        left.left(),
                        black(right.key, right.getValue(), app, right.right()));
            }
        }
    }

    static <K, V> Node<K, V> balanceLeftDel(K key, V val, Node<K, V> del, Node<K, V> right) {
        if (del instanceof Red<?, ?>) {
            return red(key, val, del.blacken(), right);
        }
        else if (right instanceof Black<?, ?>) {
            return rightBalance(key, val, del, right.redden());
        }
        else if (right instanceof Red<?, ?> && right.left() instanceof Black<?, ?>) {
            return red(
                    right.left().key,
                    right.left().getValue(),
                    black(key, val, del, right.left().left()),
                    rightBalance(right.key, right.getValue(), right.left().right(), right.right().redden()));
        }
        else {
            throw new UnsupportedOperationException("Invariant violation");
        }
    }

    static <K, V> Node<K, V> balanceRightDel(K key, V val, Node<K, V> left, Node<K, V> del) {
        if (del instanceof Red<?, ?>) {
            return red(key, val, left, del.blacken());
        }
        else if (left instanceof Black<?, ?>) {
            return leftBalance(key, val, left.redden(), del);
        }
        else if (left instanceof Red<?, ?> && left.right() instanceof Black<?, ?>) {
            return red(
                    left.right().key,
                    left.right().getValue(),
                    leftBalance(left.key, left.getValue(), left.left().redden(), left.right().left()),
                    black(key, val, left.right().right(), del));
        }
        else {
            throw new UnsupportedOperationException("Invariant violation");
        }
    }

    static <K, V> Node<K, V> leftBalance(K key, V val, Node<K, V> ins, Node<K, V> right) {
        if (ins instanceof Red<?, ?> && ins.left() instanceof Red<?, ?>) {
            return red(ins.key, ins.getValue(), ins.left().blacken(), black(key, val, ins.right(), right));
        }
        else if (ins instanceof Red<?, ?> && ins.right() instanceof Red<?, ?>) {
            return red(
                    ins.right().key,
                    ins.right().getValue(),
                    black(ins.key, ins.getValue(), ins.left(), ins.right().left()),
                    black(key, val, ins.right().right(), right));
        }
        else {
            return black(key, val, ins, right);
        }
    }

    static <K, V> Node<K, V> rightBalance(K key, V val, Node<K, V> left, Node<K, V> ins) {
        if (ins instanceof Red<?, ?> && ins.right() instanceof Red<?, ?>) {
            return red(ins.key, ins.getValue(), black(key, val, left, ins.left()), ins.right().blacken());
        }
        else if (ins instanceof Red<?, ?> && ins.left() instanceof Red<?, ?>) {
            return red(
                    ins.left().key,
                    ins.left().getValue(),
                    black(key, val, left, ins.left().left()),
                    black(ins.key, ins.getValue(), ins.left().right(), ins.right()));
        }
        else {
            return black(key, val, left, ins);
        }
    }

    Node<K, V> replace(Node<K, V> t, K key, V val) {
        int c = compare(key, t.key);
        return t.replace(t.key, c == 0 ? val : t.getValue(), c < 0 ? replace(t.left(), key, val) : t.left(),
                c > 0 ? replace(t.right(), key, val) : t.right());
    }

    static <K, V> Red<K, V> red(K key, V val, Node<K, V> left, Node<K, V> right) {
        if (left == null && right == null) {
            if (val == null) {
                return new Red<K, V>(key);
            }
            return new RedVal<K, V>(key, val);
        }
        if (val == null) {
            return new RedBranch<K, V>(key, left, right);
        }
        return new RedBranchVal<K, V>(key, val, left, right);
    }

    static <K, V> Black<K, V> black(K key, V val, Node<K, V> left, Node<K, V> right) {
        if (left == null && right == null) {
            if (val == null) {
                return new Black<K, V>(key);
            }
            return new BlackVal<K, V>(key, val);
        }
        if (val == null) {
            return new BlackBranch<K, V>(key, left, right);
        }
        return new BlackBranchVal<K, V>(key, val, left, right);
    }

    static abstract class Node<K, V> implements Map.Entry<K, V> {
        final K key;

        Node(K key) {
            this.key = key;
        }

        public K getKey() {
            return key;
        }

        public V getValue() {
            return null;
        }

        Node<K, V> left() {
            return null;
        }

        Node<K, V> right() {
            return null;
        }

        public V setValue(V value) {
            throw new UnsupportedOperationException();
        }

        abstract Node<K, V> addLeft(Node<K, V> ins);

        abstract Node<K, V> addRight(Node<K, V> ins);

        abstract Node<K, V> removeLeft(Node<K, V> del);

        abstract Node<K, V> removeRight(Node<K, V> del);

        abstract Node<K, V> blacken();

        abstract Node<K, V> redden();

        Node<K, V> balanceLeft(Node<K, V> parent) {
            return black(parent.key, parent.getValue(), this, parent.right());
        }

        Node<K, V> balanceRight(Node<K, V> parent) {
            return black(parent.key, parent.getValue(), parent.left(), this);
        }

        abstract Node<K, V> replace(K key, V val, Node<K, V> left, Node<K, V> right);

    }

    static class Black<K, V> extends Node<K, V> {
        public Black(K key) {
            super(key);
        }

        @Override
        Node<K, V> addLeft(Node<K, V> ins) {
            return ins.balanceLeft(this);
        }

        @Override
        Node<K, V> addRight(Node<K, V> ins) {
            return ins.balanceRight(this);
        }

        @Override
        Node<K, V> removeLeft(Node<K, V> del) {
            return balanceLeftDel(key, getValue(), del, right());
        }

        @Override
        Node<K, V> removeRight(Node<K, V> del) {
            return balanceRightDel(key, getValue(), left(), del);
        }

        @Override
        Node<K, V> blacken() {
            return this;
        }

        @Override
        Node<K, V> redden() {
            return new Red<K, V>(key);
        }

        @Override
        Node<K, V> replace(K key, V val, Node<K, V> left, Node<K, V> right) {
            return black(key, val, left, right);
        }
    }

    static class BlackVal<K, V> extends Black<K, V> {
        final V val;

        public BlackVal(K key, V val) {
            super(key);
            this.val = val;
        }

        @Override
        public V getValue() {
            return val;
        }

        @Override
        Node<K, V> redden() {
            return new RedVal<K, V>(key, val);
        }

    }

    static class BlackBranch<K, V> extends Black<K, V> {
        final Node<K, V> left;

        final Node<K, V> right;

        public BlackBranch(K key, Node<K, V> left, Node<K, V> right) {
            super(key);
            this.left = left;
            this.right = right;
        }

        @Override
        public Node<K, V> left() {
            return left;
        }

        @Override
        public Node<K, V> right() {
            return right;
        }

        @Override
        Node<K, V> redden() {
            return new RedBranch<K, V>(key, left, right);
        }

    }

    static class BlackBranchVal<K, V> extends BlackBranch<K, V> {
        final V val;

        public BlackBranchVal(K key, V val, Node<K, V> left, Node<K, V> right) {
            super(key, left, right);
            this.val = val;
        }

        @Override
        public V getValue() {
            return val;
        }

        @Override
        Node<K, V> redden() {
            return new RedBranchVal<K, V>(key, val, left, right);
        }

    }

    static class Red<K, V> extends Node<K, V> {
        public Red(K key) {
            super(key);
        }

        @Override
        Node<K, V> addLeft(Node<K, V> ins) {
            return red(key, getValue(), ins, right());
        }

        @Override
        Node<K, V> addRight(Node<K, V> ins) {
            return red(key, getValue(), left(), ins);
        }

        @Override
        Node<K, V> removeLeft(Node<K, V> del) {
            return red(key, getValue(), del, right());
        }

        @Override
        Node<K, V> removeRight(Node<K, V> del) {
            return red(key, getValue(), left(), del);
        }

        @Override
        Node<K, V> blacken() {
            return new Black<K, V>(key);
        }

        @Override
        Node<K, V> redden() {
            throw new UnsupportedOperationException("Invariant violation");
        }

        @Override
        Node<K, V> replace(K key, V val, Node<K, V> left, Node<K, V> right) {
            return red(key, val, left, right);
        }

    }

    static class RedVal<K, V> extends Red<K, V> {
        final V val;

        public RedVal(K key, V val) {
            super(key);
            this.val = val;
        }

        @Override
        public V getValue() {
            return val;
        }

        @Override
        Node<K, V> blacken() {
            return new BlackVal<K, V>(key, val);
        }

    }

    static class RedBranch<K, V> extends Red<K, V> {
        final Node<K, V> left;

        final Node<K, V> right;

        public RedBranch(K key, Node<K, V> left, Node<K, V> right) {
            super(key);
            this.left = left;
            this.right = right;
        }

        @Override
        public Node<K, V> left() {
            return left;
        }

        @Override
        public Node<K, V> right() {
            return right;
        }

        @Override
        Node<K, V> balanceLeft(Node<K, V> parent) {
            if (left instanceof Red<?, ?>) {
                return red(
                        key,
                        getValue(),
                        left.blacken(),
                        black(parent.key, parent.getValue(), right, parent.right()));
            }
            else if (right instanceof Red<?, ?>) {
                return red(
                        right.key,
                        right.getValue(),
                        black(key, getValue(), left, right.left()),
                        black(parent.key, parent.getValue(), right.right(), parent.right()));
            }
            else {
                return super.balanceLeft(parent);
            }

        }

        @Override
        Node<K, V> balanceRight(Node<K, V> parent) {
            if (right instanceof Red<?, ?>) {
                return red(
                        key,
                        getValue(),
                        black(parent.key, parent.getValue(), parent.left(), left),
                        right.blacken());
            }
            else if (left instanceof Red<?, ?>) {
                return red(
                        left.key,
                        left.getValue(),
                        black(parent.key, parent.getValue(), parent.left(), left.left()),
                        black(key, getValue(), left.right(), right));
            }
            else {
                return super.balanceRight(parent);
            }
        }

        @Override
        Node<K, V> blacken() {
            return new BlackBranch<K, V>(key, left, right);
        }

    }

    static class RedBranchVal<K, V> extends RedBranch<K, V> {
        final V val;

        public RedBranchVal(K key, V val, Node<K, V> left, Node<K, V> right) {
            super(key, left, right);
            this.val = val;
        }

        @Override
        public V getValue() {
            return val;
        }

        @Override
        Node<K, V> blacken() {
            return new BlackBranchVal<K, V>(key, val, left, right);
        }
    }

    static class Seq<K, V> implements ISeq<Map.Entry<K, V>> {
        final IImmutableList<Map.Entry<K, V>> stack;
        final boolean asc;
        final int cnt;

        static <K, V> Seq<K, V> create(Node<K, V> t, boolean asc, int cnt) {
            return new Seq<K, V>(push(t, ImmutableList.<Map.Entry<K, V>>empty(), asc), asc, cnt);
        }

        static <K, V> IImmutableList<Map.Entry<K, V>> push(Node<K, V> t, IImmutableList<Map.Entry<K, V>> stack, boolean asc) {
            while (t != null) {
                if (stack == null ) {
                    stack = ImmutableList.<Map.Entry<K, V>>empty();
                }
                stack = stack.with(t);
                t = asc ? t.left() : t.right();
            }
            return stack;
        }

        Seq(IImmutableList<Map.Entry<K, V>> stack, boolean asc) {
            this.stack = stack;
            this.asc = asc;
            cnt = -1;
        }

        Seq(IImmutableList<Map.Entry<K, V>> stack, boolean asc, int cnt) {
            this.stack = stack;
            this.asc = asc;
            this.cnt = cnt;
        }

        public Map.Entry<K, V> first() {
            return stack.first();
        }

        public ISeq<Map.Entry<K, V>> next() {
            Node<K, V> t = (Node<K, V>) stack.first();
            IImmutableList<Map.Entry<K, V>> nextstack = push(asc ? t.right() : t.left(), stack.rest(), asc);
            if (nextstack != null) {
                return new Seq<K, V>(nextstack, asc, cnt - 1);
            }
            return null;
        }

        public int size() {
            return cnt;
        }
    }
}
