/**
 Copyright (c) 2007-2008, Rich Hickey
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 * Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.

 * Neither the name of Clojure nor the names of its contributors
   may be used to endorse or promote products derived from this
   software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.
 **/

/* rich Jul 5, 2007 */

package com.urbancode.commons.util.immutable;

import java.util.List;

@SuppressWarnings("unchecked")
public class ImmutableVector<T> extends AImmutableVector<T> {
    final int cnt;
    final int shift;
    final Object[] root;
    final Object[] tail;

    final static private ImmutableVector<?> EMPTY =
        new ImmutableVector<Object>(0, 5, Util.EMPTY_ARRAY, Util.EMPTY_ARRAY);

    static public <T> ImmutableVector<T> empty() {
        return (ImmutableVector<T>) EMPTY;
    }

    static public <T> ImmutableVector<T> create(List<? extends T> items) {
        ImmutableVector<T> ret = empty();
        for (T item : items) {
            ret = ret.with(item);
        }
        return ret;
    }

    static public <T> ImmutableVector<T> create(T... items) {
        ImmutableVector<T> ret = empty();
        for (T item : items) {
            ret = ret.with(item);
        }
        return ret;
    }

    ImmutableVector(int cnt, int shift, Object[] root, Object[] tail) {
        this.cnt = cnt;
        this.shift = shift;
        this.root = root;
        this.tail = tail;
    }

    public T get(int i) {
        if (i >= 0 && i < cnt) {
            if (i >= tailOffset()) {
                return (T) tail[i & 0x01f];
            }
            Object[] arr = root;
            for (int level = shift; level > 0; level -= 5) {
                arr = (Object[]) arr[(i >>> level) & 0x01f];
            }
            return (T) arr[i & 0x01f];
        }
        throw new IndexOutOfBoundsException();
    }

    public ImmutableVector<T> with(int i, T val) {
        if (i >= 0 && i < cnt) {
            if (i >= tailOffset()) {
                Object[] newTail = new Object[tail.length];
                System.arraycopy(tail, 0, newTail, 0, tail.length);
                newTail[i & 0x01f] = val;

                return new ImmutableVector<T>(cnt, shift, root, newTail);
            }

            return new ImmutableVector<T>(cnt, shift, doAssoc(shift, root, i, val), tail);
        }
        if (i == cnt) {
            return with(val);
        }
        throw new IndexOutOfBoundsException();
    }

    public int size() {
        return cnt;
    }

    public ImmutableVector<T> with(T val) {
        if (tail.length < 32) {
            Object[] newTail = new Object[tail.length + 1];
            System.arraycopy(tail, 0, newTail, 0, tail.length);
            newTail[tail.length] = val;
            return new ImmutableVector<T>(cnt + 1, shift, root, newTail);
        }
        Box expansion = new Box(null);
        Object[] newRoot = pushTail(shift - 5, root, tail, expansion);
        int newShift = shift;
        if (expansion.val != null) {
            newRoot = new Object[] { newRoot, expansion.val };
            newShift += 5;
        }
        return new ImmutableVector<T>(cnt + 1, newShift, newRoot, new Object[] { val });
    }

    public ImmutableVector<T> pop() {
        if (cnt == 0) {
            throw new IllegalStateException("Can't pop empty vector");
        }
        if (cnt == 1) {
            return empty();
        }
        if (tail.length > 1) {
            Object[] newTail = new Object[tail.length - 1];
            System.arraycopy(tail, 0, newTail, 0, newTail.length);
            return new ImmutableVector<T>(cnt - 1, shift, root, newTail);
        }
        Box ptail = new Box(null);
        Object[] newroot = popTail(shift - 5, root, ptail);
        int newshift = shift;
        if (newroot == null) {
            newroot = Util.EMPTY_ARRAY;
        }
        if (shift > 5 && newroot.length == 1) {
            newroot = (Object[]) newroot[0];
            newshift -= 5;
        }
        return new ImmutableVector<T>(cnt - 1, newshift, newroot, (Object[]) ptail.val);
    }

    private int tailOffset() {
        return cnt - tail.length;
    }

    private Object[] pushTail(int level, Object[] arr, Object[] tailNode, Box expansion) {
        Object newchild;
        if (level == 0) {
            newchild = tailNode;
        }
        else {
            newchild = pushTail(level - 5, (Object[]) arr[arr.length - 1], tailNode, expansion);
            if (expansion.val == null) {
                Object[] ret = arr.clone();
                ret[arr.length - 1] = newchild;
                return ret;
            }
            else {
                newchild = expansion.val;
            }
        }
        // expansion
        if (arr.length == 32) {
            expansion.val = new Object[] { newchild };
            return arr;
        }
        Object[] ret = new Object[arr.length + 1];
        System.arraycopy(arr, 0, ret, 0, arr.length);
        ret[arr.length] = newchild;
        expansion.val = null;
        return ret;
    }

    private Object[] popTail(int shift, Object[] arr, Box ptail) {
        if (shift > 0) {
            Object[] newchild = popTail(shift - 5, (Object[]) arr[arr.length - 1], ptail);
            if (newchild != null) {
                Object[] ret = arr.clone();
                ret[arr.length - 1] = newchild;
                return ret;
            }
        }
        if (shift == 0) {
            ptail.val = arr[arr.length - 1];
        }
        // contraction
        if (arr.length == 1) {
            return null;
        }
        Object[] ret = new Object[arr.length - 1];
        System.arraycopy(arr, 0, ret, 0, ret.length);
        return ret;
    }

    private static Object[] doAssoc(int level, Object[] arr, int i, Object val) {
        Object[] ret = arr.clone();
        if (level == 0) {
            ret[i & 0x01f] = val;
        }
        else {
            int subidx = (i >>> level) & 0x01f;
            ret[subidx] = doAssoc(level - 5, (Object[]) arr[subidx], i, val);
        }
        return ret;
    }
}
